/*
 * Copyright 2015-2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the OpenSSL license (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#include <openssl/ec.h>
#include "ec_lcl.h"
#include <openssl/err.h>
#ifdef NETAPP
# include <openssl/evp.h>
# include "../evp/evp_locl.h"
#endif /* NETAPP */

ECDSA_SIG *ECDSA_do_sign(const unsigned char *dgst, int dlen, EC_KEY *eckey)
{
    return ECDSA_do_sign_ex(dgst, dlen, NULL, NULL, eckey);
}

ECDSA_SIG *ECDSA_do_sign_ex(const unsigned char *dgst, int dlen,
                            const BIGNUM *kinv, const BIGNUM *rp,
                            EC_KEY *eckey)
{
    if (eckey->meth->sign_sig != NULL)
        return eckey->meth->sign_sig(dgst, dlen, kinv, rp, eckey);
    ECerr(EC_F_ECDSA_DO_SIGN_EX, EC_R_OPERATION_NOT_SUPPORTED);
    return NULL;
}

int ECDSA_sign(int type, const unsigned char *dgst, int dlen, unsigned char
               *sig, unsigned int *siglen, EC_KEY *eckey)
{
    return ECDSA_sign_ex(type, dgst, dlen, sig, siglen, NULL, NULL, eckey);
}

int ECDSA_sign_ex(int type, const unsigned char *dgst, int dlen,
                  unsigned char *sig, unsigned int *siglen, const BIGNUM *kinv,
                  const BIGNUM *r, EC_KEY *eckey)
{
#ifdef NETAPP
#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        const EVP_MD *evp_md;
        // The following call will return a valid digest structure only if the application
        // has added the digest we are trying to lookup. The digest could have been added
        // with a call to EVP_add_digest() or OpenSSL_add_all_digests() prior to this.
        evp_md = EVP_get_digestbynid(type);
        if (evp_md) {
            evp_md = evp_get_fips_md(evp_md);
        }
        if ((evp_md && !(evp_md->flags & EVP_MD_FLAG_FIPS)) || (type == NID_sha1)) {
            ECerr(EC_F_ECDSA_SIGN_EX, EC_R_FIPS_DISALLOWED_MD_FOR_SIGN);
            return 0;
        }
        // The following code to restrict EC key size for signature generation
        // will be removed after the same restriction is put in the FIPS module(NCSM)
        const EC_GROUP *group;
        BIGNUM *order = NULL;
        if ((order=BN_new()) && (group = EC_KEY_get0_group(eckey)) && (EC_GROUP_get_order(group, order, NULL))) {
             if (BN_num_bits(order) < OPENSSL_EC_FIPS_MIN_SIGN_BITS) {
                 ECerr(EC_F_ECDSA_SIGN_EX, EC_R_FIPS_KEY_TOO_SHORT);
                 BN_free(order);
                 return 0;
             }
        }
        if (order) {
            BN_free(order);
        }
    }
#endif
#endif /* NETAPP */
    if (eckey->meth->sign != NULL)
        return eckey->meth->sign(type, dgst, dlen, sig, siglen, kinv, r, eckey);
    ECerr(EC_F_ECDSA_SIGN_EX, EC_R_OPERATION_NOT_SUPPORTED);
    return 0;
}

int ECDSA_sign_setup(EC_KEY *eckey, BN_CTX *ctx_in, BIGNUM **kinvp,
                     BIGNUM **rp)
{
    if (eckey->meth->sign_setup != NULL)
        return eckey->meth->sign_setup(eckey, ctx_in, kinvp, rp);
    ECerr(EC_F_ECDSA_SIGN_SETUP, EC_R_OPERATION_NOT_SUPPORTED);
    return 0;
}
