/*
 * Copyright 2002-2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the OpenSSL license (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#include <openssl/ec.h>
#include "ec_lcl.h"
#include <openssl/err.h>

/*-
 * returns
 *      1: correct signature
 *      0: incorrect signature
 *     -1: error
 */
int ECDSA_do_verify(const unsigned char *dgst, int dgst_len,
                    const ECDSA_SIG *sig, EC_KEY *eckey)
{
#ifdef NETAPP
#ifdef OPENSSL_FIPS
    if (FIPS_mode()) {
        // The following code to restrict EC key size for signature verification
        // will be removed after the same restriction is put in the FIPS module(NCSM).
        const EC_GROUP *group;
        BIGNUM *order = NULL;
        if ((order=BN_new()) && (group = EC_KEY_get0_group(eckey)) && (EC_GROUP_get_order(group, order, NULL))) {
            if (BN_num_bits(order) < OPENSSL_EC_FIPS_MIN_VERIFY_BITS) {
                ECerr(EC_F_ECDSA_DO_VERIFY, EC_R_FIPS_KEY_TOO_SHORT);
                BN_free(order);
                return -1;
            }
        }
        if (order) {
            BN_free(order);
        }
    }
#endif
#endif /* NETAPP */
    if (eckey->meth->verify_sig != NULL)
        return eckey->meth->verify_sig(dgst, dgst_len, sig, eckey);
    ECerr(EC_F_ECDSA_DO_VERIFY, EC_R_OPERATION_NOT_SUPPORTED);
    return 0;
}

/*-
 * returns
 *      1: correct signature
 *      0: incorrect signature
 *     -1: error
 */
int ECDSA_verify(int type, const unsigned char *dgst, int dgst_len,
                 const unsigned char *sigbuf, int sig_len, EC_KEY *eckey)
{
    if (eckey->meth->verify != NULL)
        return eckey->meth->verify(type, dgst, dgst_len, sigbuf, sig_len,
                                   eckey);
    ECerr(EC_F_ECDSA_VERIFY, EC_R_OPERATION_NOT_SUPPORTED);
    return 0;
}
