/*
 * Copyright 2012-2017 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the OpenSSL license (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#include <stdio.h>
#include <openssl/crypto.h>
#include "internal/cryptlib.h"
#include <openssl/conf.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#ifdef NETAPP
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif
void alg_module_unload(CONF_MODULE *pmod);
#endif /* NETAPP */

/* Algorithm configuration module. */

static int alg_module_init(CONF_IMODULE *md, const CONF *cnf)
{
#ifndef NETAPP
    int i;
    const char *oid_section;
    STACK_OF(CONF_VALUE) *sktmp;
    CONF_VALUE *oval;

    oid_section = CONF_imodule_get_value(md);
    if ((sktmp = NCONF_get_section(cnf, oid_section)) == NULL) {
        EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_ERROR_LOADING_SECTION);
        return 0;
    }
    for (i = 0; i < sk_CONF_VALUE_num(sktmp); i++) {
        oval = sk_CONF_VALUE_value(sktmp, i);
        if (strcmp(oval->name, "fips_mode") == 0) {
            int m;
            if (!X509V3_get_value_bool(oval, &m)) {
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INVALID_FIPS_MODE);
                return 0;
            }
            if (m > 0) {
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_FIPS_MODE_NOT_SUPPORTED);
                return 0;
            }
        } else {
            EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_UNKNOWN_OPTION);
            ERR_add_error_data(4, "name=", oval->name,
                               ", value=", oval->value);
        }

    }
    return 1;
#else /* NETAPP */
    int i;
    const char *oid_section;
    STACK_OF(CONF_VALUE) *sktmp;
    CONF_VALUE *oval;
    oid_section = CONF_imodule_get_value(md);
    void *usr_data = NULL;
    ALG_MODULE_DATA  *alg_md_data = NULL;

    /* Set the unload func in parent module */
    CONF_MODULE *pmod = CONF_imodule_get_module(md);
    if(!CONF_module_get_conf_unload_func(pmod))
        CONF_module_set_conf_unload_func(pmod, alg_module_unload);

    if ((sktmp = NCONF_get_section(cnf, oid_section)) == NULL) {
        EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_ERROR_LOADING_SECTION);
        return 0;
    }

    CRYPTO_THREAD_ID cur = CRYPTO_THREAD_get_current_id();

    for (i = 0; i < sk_CONF_VALUE_num(sktmp); i++) {
        oval = sk_CONF_VALUE_value(sktmp, i);
        if ((strcmp(oval->name, "fips_mode") == 0) ||
                (strcmp(oval->name, "ssl_protocol") == 0) ||
                (strcmp(oval->name, "ssl_ciphers") == 0)) {

            alg_conf_wlock();
            usr_data = CONF_module_get_usr_data(pmod);

            if (!usr_data) {
                usr_data = OPENSSL_malloc(sizeof(ALG_MODULE_DATA));
                if (!usr_data)
                    goto err;
                memset(usr_data, 0, sizeof(ALG_MODULE_DATA));

                CONF_module_set_usr_data(pmod, usr_data);
            }

            alg_md_data = (ALG_MODULE_DATA *) usr_data;

            // Set current thread's id
            if(alg_md_data) {
                if (CRYPTO_THREAD_compare_id(cur, alg_md_data->thr_id)) {
                    alg_md_data->thr_id = cur;
                }
            }

            if (strcmp(oval->name, "fips_mode") == 0) {
                int m;
                if (!X509V3_get_value_bool(oval, &m)) {
                    EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INVALID_FIPS_MODE);
                    goto err;
                }
                alg_md_data->fips_mode = m;

                /* IMP: We must unlock CRYPTO_LOCK_ALG_CONF before invoking
                   FIPS_mode_set(1) */
                alg_conf_unlock();

                if (m > 0) {
#ifdef OPENSSL_FIPS
                    if (!FIPS_module_mode() && !FIPS_mode_set(1)) {
                        EVPerr(EVP_F_ALG_MODULE_INIT,
                            EVP_R_ERROR_SETTING_FIPS_MODE);
                        alg_conf_wlock();
                        goto err;
                    }
#else
                    EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_FIPS_MODE_NOT_SUPPORTED);
                    alg_conf_wlock();
                    goto err;
#endif
                }

                /* NOTE: We must continue here to avoid having to unlock the lock */
                continue;

            } /* fips_mode */
            else if (strcmp(oval->name, "ssl_protocol") == 0) {
                if(alg_md_data->ssl_protocol_list)
                    goto done;
                alg_md_data->ssl_protocol_list = X509V3_parse_list(oval->value);
                if (!alg_md_data->ssl_protocol_list)
                    goto err;
            } /* ssl_protocol */
            else if (strcmp(oval->name, "ssl_ciphers") == 0) {
                if(alg_md_data->ssl_cipher_list)
                    goto done;
                alg_md_data->ssl_cipher_list = OPENSSL_strdup(oval->value);
                if (!alg_md_data->ssl_cipher_list)
                    goto err;
            } /* ssl_ciphers */

            done:
                alg_conf_unlock();
                continue;

            err:
                if (alg_md_data) {
                    if (alg_md_data->ssl_protocol_list)
                        sk_CONF_VALUE_pop_free(alg_md_data->ssl_protocol_list, X509V3_conf_free);
                    if (alg_md_data->ssl_cipher_list)
                        OPENSSL_free(alg_md_data->ssl_cipher_list);
                    OPENSSL_free(alg_md_data);
                    CONF_module_set_usr_data(pmod, NULL);
                }
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INSUFFICIENT_MEMORY);
                alg_conf_unlock();
                return 0;

        } else {
            EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_UNKNOWN_OPTION);
            ERR_add_error_data(4, "name=", oval->name,
                               ", value=", oval->value);
        }
    }
    return 1;
#endif /* NETAPP */
}

#ifdef NETAPP
void alg_module_unload(CONF_MODULE *pmod)
{
    alg_conf_wlock();
    void *usr_data = CONF_module_get_usr_data(pmod);
    ALG_MODULE_DATA  *alg_md_data = (ALG_MODULE_DATA *) usr_data;
    if (alg_md_data) {
        if (alg_md_data->ssl_protocol_list)
            sk_CONF_VALUE_pop_free(alg_md_data->ssl_protocol_list, X509V3_conf_free);
        if (alg_md_data->ssl_cipher_list)
            OPENSSL_free(alg_md_data->ssl_cipher_list);
        OPENSSL_free(alg_md_data);
        CONF_module_set_usr_data(pmod, NULL);
    }
    alg_conf_unlock();
}
#endif /* NETAPP */

void EVP_add_alg_module(void)
{
    CONF_module_add("alg_section", alg_module_init, 0);
}
