/*
 * Copyright 2011-2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the OpenSSL license (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#include "internal/cryptlib.h"
#ifdef NETAPP
#include <openssl/conf.h>
#include <openssl/evp.h>
#include <openssl/err.h>

#include <internal/thread_once.h>

static CRYPTO_RWLOCK *fips_init_lock = NULL;

static CRYPTO_ONCE fips_lock_init = CRYPTO_ONCE_STATIC_INIT;

DEFINE_RUN_ONCE_STATIC(do_fips_lock_init)
{
    fips_init_lock = CRYPTO_THREAD_lock_new();
    return fips_init_lock != NULL;
}

void fips_i_lock(void)
{
    (void) RUN_ONCE(&fips_lock_init, do_fips_lock_init);
    CRYPTO_THREAD_write_lock(fips_init_lock);
}

void fips_i_unlock(void)
{
    (void) CRYPTO_THREAD_unlock(fips_init_lock);
}

/* Perform any essential FIPS initialization operations.
 * Currently reads and applies configuration
 *
 * NOTE: This routine is invoked currently from these APIs:
 * EVP_add_cipher(), EVP_add_digest(), FIPS_mode() and
 * FIPS_mode_set()
 *
 * Adding it in EVP_add_cipher/digest ensures the library
 * initialization routines like OpenSSL_add_all_algorithms()
 * and SSL_library_init() are covered. Adding it in FIPS_mode*
 * APIs ensures OpenSSL to OpenSSL-FIPS transition happens
 * only after initialization of the FIPS mode as per the
 * configuration. Although, transition from OpenSSL library
 * calls to FIPS-canister contained CRYPTO calls should happen
 * after checking of FIPS_mode(), we may have to revisit this
 * if the OpenSSL design is changed to do such transition
 * without of explicit check of FIPS_mode() as such.
 *
 */
void FIPS_init(void)
{
    if (get_OPENSSL_init_done()) {
        CONF_MODULE *md = NULL;
        ALG_MODULE_DATA *alg_data = NULL;
        CRYPTO_THREAD_ID cur = CRYPTO_THREAD_get_current_id();

        // So the obvious optimization is that we need to avoid doing the
        // significant optimization work every time through here. Unfortunately,
        // we can't just assume once is enough since the config could be
        // unloaded without unloading the whole library. So, we have to check
        // the alg_data thing to convince ourselves a config really is there.
        // Even more unfortunately, the CONF_modules_load_file chain will end
        // up calling into here recursively, so we have to do some recursion
        // detection to avoid eventual deadlock there.

        alg_conf_rlock();
        if ((md = module_find("alg_section")) &&
            (alg_data = (ALG_MODULE_DATA *) CONF_module_get_usr_data(md))) {
            if (!CRYPTO_THREAD_compare_id(alg_data->thr_id, cur)) {
                alg_conf_unlock();
                return;
            }
        }
        alg_conf_unlock();

        fips_i_lock();

        alg_conf_rlock();
        if ((md = module_find("alg_section")) &&
            (alg_data = (ALG_MODULE_DATA *) CONF_module_get_usr_data(md)) &&
            alg_data->init_done) {
            alg_conf_unlock();
            fips_i_unlock();
            return;
        }
        alg_conf_unlock();

        alg_conf_wlock();
        EVP_add_alg_module();
        alg_conf_unlock();
        ERR_clear_error();
        CONF_modules_load_file(NULL /* default config file */,
                               NULL /* default section */,
                               CONF_MFLAGS_DEFAULT_SECTION |
                               CONF_MFLAGS_IGNORE_MISSING_FILE);

        alg_conf_wlock();
        if ((md = module_find("alg_section")) &&
            (alg_data = (ALG_MODULE_DATA *) CONF_module_get_usr_data(md))) {
            alg_data->init_done = 1;
        }
        alg_conf_unlock();

        fips_i_unlock();
    }
}
#endif  /* NETAPP */


int FIPS_mode(void)
{
#ifdef NETAPP
    OPENSSL_init();
    FIPS_init();
#endif /* NETAPP */
    /* This version of the library does not support FIPS mode. */
    return 0;
}

int FIPS_mode_set(int r)
{
#ifdef NETAPP
   OPENSSL_init();
#ifdef OPENSSL_FIPS
    if ( r == 0 )
    {
        /* Check the configuration and disallow only if the
         * fips_mode=yes string is present in configuration
         */
        CONF_MODULE *md = NULL;
        ALG_MODULE_DATA *alg_data = NULL;

        alg_conf_rlock();
        if ((md = module_find("alg_section")) &&
            (alg_data = (ALG_MODULE_DATA *) CONF_module_get_usr_data(md)) &&
            (alg_data->fips_mode))
        {
            CRYPTOerr(CRYPTO_F_FIPS_MODE_SET, CRYPTO_R_FIPS_MODE_UNSET_NOT_ALLOWED);
            alg_conf_unlock();
            return 0;
        }
        alg_conf_unlock();
    }
#endif
#endif /* NETAPP */
    if (r == 0)
        return 1;
    CRYPTOerr(CRYPTO_F_FIPS_MODE_SET, CRYPTO_R_FIPS_MODE_NOT_SUPPORTED);
    return 0;
}
